package badger

import (
	"fmt"
	"log"
	"os"
	"time"
)

var backupTicker *time.Ticker

func setupBackup(db *Badger, interval int) error {
	if interval > 0 {
		log.Printf("periodical database backup interval %v seconds", interval)
		if _, err := os.Stat("backup"); os.IsNotExist(err) {
			err = os.Mkdir("backup", 0700)
			if err != nil {
				return err
			}
		}
		backupTicker = time.NewTicker(time.Duration(interval) * time.Second)
		go func() {
			for {
				select {
				case <-backupTicker.C:
					var intermediate *os.File
					c := func() {
						if err := intermediate.Close(); err != nil {
							log.Printf("error closing intermediate backup file: %s", err)
						}
					}

					if i, err := os.Create("backup/.intermediate"); err != nil {
						log.Printf("error creating intermediate backup file: %s", err)
						continue
					} else {
						intermediate = i
					}

					var ver uint64
					if v, err := db.DB.Backup(intermediate, 0); err != nil {
						log.Printf("error doing backup: %s", err)
						c()
						continue
					} else {
						ver = v
					}

					if err := os.Rename("backup/.intermediate", fmt.Sprintf("backup/%v", ver)); err != nil {
						log.Printf("error renaming intermediate backup file: %s", err)
						if err = os.Remove("backup/.intermediate"); err != nil {
							log.Printf("error removing intermediate file: %s", err)
							log.Printf("backup has been disabled, see above two messages")
							c()
							break
						}
						c()
						continue
					}
					c()
					log.Printf("database backup successful, version %v", ver)
				}
			}
		}()
		return nil
	} else {
		log.Printf("periodical database backup disabled")
		return nil
	}
}
